import ctypes
from ..native.Python.QuixSdkStreaming.Models.StreamWriter.EventDefinitionBuilder import EventDefinitionBuilder as edbi

from ..helpers.enumconverter import EnumConverter as ec
from ..models.eventlevel import EventLevel

from ..native.Python.QuixSdkProcess.Models.EventLevel import EventLevel as InteropEventLevel
from ..helpers.nativedecorator import nativedecorator


@nativedecorator
class EventDefinitionBuilder(object):
    """
        Builder for creating event and group definitions for StreamPropertiesWriter
    """

    def __init__(self, net_pointer: ctypes.c_void_p):
        """
            Initializes a new instance of EventDefinitionBuilder.

            Parameters:

            net_pointer: Pointer to an instance of a .net EventDefinitionBuilder.
        """

        if net_pointer is None:
            raise Exception("EventDefinitionBuilder is none")

        self.__interop = edbi(net_pointer)

    def set_level(self, level: EventLevel) -> 'EventDefinitionBuilder':
        """
        Set the custom properties of the event
        :param level: The severity level of the event
        :return: The builder
        """

        converted_level = ec.enum_to_another(level, InteropEventLevel)
        new = edbi(self.__interop.SetLevel(converted_level))
        if new != self.__interop:
            self.__interop._dispose_ptr()
            self.__interop = new
        return self

    def set_custom_properties(self, custom_properties: str) -> 'EventDefinitionBuilder':
        """
        Set the custom properties of the event
        :param custom_properties: The custom properties of the event
        :return: The builder
        """

        new = edbi(self.__interop.SetCustomProperties(custom_properties))
        if new != self.__interop:
            self.__interop._dispose_ptr()
            self.__interop = new
        return self

    def add_definition(self, event_id: str, name: str = None, description: str = None) -> 'EventDefinitionBuilder':
        """
        Add new event definition to the StreamPropertiesWriter. Configure it with the builder methods.
        :param event_id: The id of the event. Must match the event id used to send data.
        :param name: The human friendly display name of the event
        :param description: The description of the event
        :return: The builder
        """

        new = edbi(self.__interop.AddDefinition(event_id, name, description))
        if new != self.__interop:
            self.__interop._dispose_ptr()
            self.__interop = new
        return self
