from typing import Union, Dict
import ctypes
from ..native.Python.QuixSdkStreaming.Models.StreamWriter.ParameterDataBuilder import ParameterDataBuilder as pdbi
from ..helpers.nativedecorator import nativedecorator


@nativedecorator
class ParameterDataBuilder(object):
    """
        Builder for creating parameter data packages for StreamPropertiesWriter
    """

    def __init__(self, net_pointer: ctypes.c_void_p):
        """
            Initializes a new instance of ParameterDataBuilder.

            Parameters:

            net_pointer: Pointer to an instance of a .net ParameterDataBuilder.
        """

        if net_pointer is None:
            raise Exception("ParameterDataBuilder is none")

        self.__interop = pdbi(net_pointer)
        self.__entered = False

    def __enter__(self):
        self.__entered = True

    def add_value(self, parameter_id: str, value: Union[str, float, int, bytes, bytearray]) -> 'ParameterDataBuilder':
        """
        Adds new parameter value at the time the builder is created for
        :param parameter_id: The id of the parameter to set the value for
        :param value: the value as string or float
        :return: The builder
        """

        val_type = type(value)
        if val_type is int:
            value = float(value)
            val_type = float
        elif val_type is bytearray:
            value = bytes(value)
            val_type = bytes

        if val_type is float:
            new = pdbi(self.__interop.AddValue(parameter_id, value))
            if new != self.__interop:
                self.__interop._dispose_ptr()
                self.__interop = new
        elif val_type is str:
            new = pdbi(self.__interop.AddValue2(parameter_id, value))
            if new != self.__interop:
                self.__interop._dispose_ptr()
                self.__interop = new
        elif val_type is bytes:
            # TODO fix binary
            # new = pdbi(self.__interop.AddValue3(parameter_id, value))
            #if new != self.__interop:
             #   self.__interop._dispose_ptr()
#                self.__interop = new
            pass
        else:
            raise Exception("Invalid type " + str(val_type) + " passed as parameter value.")
        return self

    def add_tag(self, tag_id: str, value: str) -> 'ParameterDataBuilder':
        """
        Adds tag value for the values. If
        :param tag_id: The id of the tag
        :param value: The value of the tag
        :return: The builder
        """
        new = pdbi(self.__interop.AddTag(tag_id, value))
        if new != self.__interop:
            self.__interop._dispose_ptr()
            self.__interop = new
        return self

    def add_tags(self, tags: Dict[str, str]) -> 'EventDataBuilder':
        """
            Copies the tags from the specified dictionary. Conflicting tags will be overwritten
            :param tags: The tags to add

        :return: EventDataBuilder
        """

        if tags is None:
            return self
        # TODO
        prep_tags_dict = System.Collections.Generic.List[System.Collections.Generic.KeyValuePair[str, str]]([])
        for key, val in tags.items():
            prep_tags_dict.Add(System.Collections.Generic.KeyValuePair[str, str](key, val))
        self.__wrapped.AddTags(prep_tags_dict)
        return self

    def write(self):
        """
        Writes the values to the StreamParametersWriter buffer. See StreamParametersWriter buffer settings for more information when the values are sent to the broker
        """

        self.__interop.Write()

        if not self.__entered:
            self.dispose()
